# -*- coding: utf-8 -*-

"""

PySSE v0.2.3

Copyright (C) 2009 Emmanuel Benoît <ebenoit@ebenoit.info>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from copy import deepcopy
from mod_python import apache
from lxml import etree
import re




##############################################################################
# UTILITY CLASSES AND FUNCTIONS                                              #
##############################################################################

class SiteXML( object ):
    """
    This class serves as a placeholder for the XML tree of the site's main
    configuration. The file is only opened and parsed when the object's xml
    property is accessed.
    """

    def __init__( self , data_path ):
        """
        Stores the file's path into the instance and mark the tree as absent
        from memory.
        """
        self.path = data_path
        self.tree = None

    @property
    def xml( self ):
        """
        This property corresponds to the XML tree for the site's main
        configuration file. If the file hadn't been opened yet, it is parsed.
        """
        if self.tree is not None:
            return self.tree
        self.tree = etree.parse( self.path + 'site.xml' )
        return self.tree

    def get( self , xpath_expr , **kwargs ):
        kwargs[ 'namespaces' ] = {
                 'ws': 'http://www.ebenoit.info/data'
            }
        return self.xml.xpath( xpath_expr , **kwargs )



class XPathFunction( object ):
    """
    This class acts as the base class for extra XPath functions provided by
    PySSE. It contains the function registration code, as well as the call
    mechanism for calls coming from either the XPath code or from other parts
    of the PySSE script. It also includes some utilities to generate XHTML
    and PySSE tags.
    
    Classes that inherit XPathFunction must have either a get_node() method
    that returns a list of elements or a get_value() method that returns a
    scalar. They may also implement a prepare() method which will be called
    during initialisation. Finally, they must have a field named function,
    which contains the name of the new XPath function.
    """

    def __init__( self , registry ):
        """
        This method registers the function in the XPath function registry and
        prepares the function by calling the class' prepare() method (if it
        exists).
        """
        self.registry = registry
        self.site = registry.site
        registry.namespace[ self.__class__.function ] = self
        if hasattr( self , 'prepare' ):
            self.prepare( )

    def __call__( self , context , *args ):
        """
        This method implements calls from the XPath code - it calls the
        internal call handling method _call() and handles exceptions
        depending on the type of the function. If the class has no handling
        method, NotImplementedError is raised.
        """
        try:
            return self._call( *args )
        except ( Exception , ) , e:
            if hasattr( self , 'get_node' ):
                return [ self.ws( 'data' , tags = [
                    self.html( 'h1' , 'Internal error' ) ,
                    self.html( 'p' , e ) ] ) ]
            elif hasattr( self , 'get_value' ):
                return u'Internal error: %s' % ( e )
            raise

    def _call( self , *args ):
        """
        This method implements actual calls to the function. It looks for a
        get_node() method, calling it if it exists; if not, it then looks for
        the get_value() method and calls it. If none of the methods exist, it
        raises the NotImplementedError exception.
        """
        if hasattr( self , 'get_node' ):
            l = self.get_node( *args )
            return [ self.ws( 'data' , tags = l ) ]
        elif hasattr( self , 'get_value' ):
            return self.get_value( *args )
        raise NotImplementedError

    def func( self , cls , *args ):
        """Calls another XPath function through the registry."""
        return self.registry.call( cls , *args )

    def html( self , tag , contents = None , tags = None ):
        """Generates an XHTML tag."""
        return self._tag( tag , contents , tags )

    def ws( self , tag , contents = None , tags = None ):
        """Generates a PySSE tag."""
        return self._tag( tag , contents , tags , 'ws' )

    def _tag( self , tag , contents , tags , ns = None ):
        """
        This method generates a tag. The 'tag' parameter determines the name
        of the new tag. If the 'contents' parameter is set, then it is used
        as the new node's text; otherwise, if the 'tags' parameters contains
        a list, it is used as the new tag's list of children. Finally, the
        'ns' parameter indicates the namespace to use; it must be None for
        XHTML tags or 'ws' for PySSE tags.
        """
        xhtml_ns = 'http://www.w3.org/1999/xhtml'
        ws_ns = 'http://www.ebenoit.info/data'
        nsmap = { None : xhtml_ns , 'ws' : ws_ns }
        node = etree.Element( '{%s}%s' % ( nsmap[ ns ] , tag ) ,
                nsmap = nsmap )
        if contents is not None:
            node.text = unicode( contents )
        elif tags is not None:
            node.extend( tags )
        return node



class XPathRegistry( object ):
    """This class implements the XPath functions registry."""

    def __init__( self , site ):
        """
        This method initialises the object by storing the XMLSite instance
        for the site, and creating both the function dictionnary and the XPath
        namespace.
        """
        self.site = site
        self.functions = {}
        self.namespace = etree.FunctionNamespace(
                'http://www.ebenoit.info/ext' )

    def add( self , *args ):
        """
        This method adds any number of functions (implemented by subclassing
        the XPathFunction class) to the registry.
        """
        for f_class in args:
            assert( f_class not in self.functions )
            f = f_class( self )
            self.functions[ f_class ] = f

    def call( self , cls , *args ):
        """This method calls an XPath function directly."""
        return self.functions[ cls ]._call( *args )



def get_languages( site ):
    """
    Returns a set of supported language identifiers. The identifiers are read
    from the main configuration file's ws:contents tags' lang attribute.
    """
    r = site.get( '/ws:data/ws:contents' )
    languages = set([ node.get( 'lang' ) for node in r ])
    if not languages:
        languages = set([ 'en' ])
    return languages



def guess_language( accept_str , languages ):
    """
    Attempts to guess the language to use from the browser's Accept-Language
    string and the set of available languages.
    """
    accept_str = re.sub( r'[^a-zA-Z0-9\.\-;,=]' , '' , accept_str )
    if len( accept_str ) == 0:
        return languages[ 0 ]

    # Build a list using the different weights as keys
    accepted = { }
    acc_list = accept_str.split( ',' )

    for lang_spec in acc_list:

        if lang_spec == '':
            continue

        lang_data = lang_spec.split( ';q=' )
        if len( lang_data ) == 1:
            lang_data.append( '1.0' )

        lang_string = lang_data[ 0 ]
        try:
            lang_weight = float( lang_data[ 1 ] ) * 1000
        except ValueError:
            continue

        if lang_weight in accepted:
            accepted[ lang_weight ].append( lang_string )
        else:
            accepted[ lang_weight ] = [ lang_string ]

    # Iterate for all languages in the weight-specific order
    weights = accepted.keys( )
    weights.sort( )
    for lang_weight in reversed( weights ) :
        for lang_string in accepted[ lang_weight ]:
            if lang_string in languages:
                return lang_string

    return languages[ 0 ]



def get_paths( req ):
    """
    Returns the various (script URI, web root, data path) from the request
    information.
    """
    # Get the script's file name
    from os.path import basename, dirname
    script_file = req.finfo[ apache.FINFO_FNAME ]

    # Get the script's URI
    script_name = basename( script_file )
    script_uri = req.parsed_uri[ apache.URI_PATH ]
    if script_uri.find( script_name ) == -1:
        script_uri += script_name
    elif req.path_info:
        script_uri = script_uri[ 0 : -len( req.path_info ) ]

    # Get the site's base URI
    base_uri = dirname( script_uri )
    if base_uri[ len( base_uri ) - 1 ] != '/' :
        base_uri = base_uri + '/'

    # Get the data path
    data_path = dirname( script_file ) + '/data/'

    return ( script_uri , base_uri , data_path )




##############################################################################
# XPATH FUNCTIONS                                                            #
##############################################################################

class GetI18NString( XPathFunction ):

    function = 'get-i18n-string'

    def prepare( self ):
        self.translations = {}
        self.translation_files = {}

    def get_value( self , string_id , lang_id ):

        # Is it cached?
        if ( string_id , lang_id ) in self.translations:
            return self.translations[ ( string_id , lang_id ) ]

        # Look in the main configuration file
        xpath_root = "/ws:data/ws:contents[ @lang = $lid ]/ws:strings"
        xpath_expr = xpath_root + "/ws:string[ @id = $sid ]"
        r = self.site.get( xpath_expr , lid = lang_id , sid = string_id )
        if len( r ) >= 1:
            self.translations[ ( string_id , lang_id ) ] = r[ 0 ].text
            return self.translations[ ( string_id , lang_id ) ]

        # Is there a separate data file?
        if lang_id not in self.translation_files:
            xpath_expr = xpath_root + "[ @datafile ]"
            r = self.site.get( xpath_expr , lid = lang_id )
            if len( r ) < 1:
                return unicode( '(string not found)' )
            self.translation_files[ lang_id ] = etree.parse(
                    self.site.path + r[0].attrib[ 'datafile' ]
                )

        # Look in the separate file
        f = self.translation_files[ lang_id ]
        xpath_expr = "/ws:strings/ws:string[ @id = $sid ]"
        r = f.xpath( xpath_expr , sid = string_id , namespaces = {
                'ws' : 'http://www.ebenoit.info/data' ,
            } )
        if len( r ) < 1:
            return unicode( '(string not found)' )
        self.translations[ ( string_id , lang_id ) ] = r[ 0 ].text
        return self.translations[ ( string_id , lang_id ) ]



class GetPageTitle( XPathFunction ):

    function = 'get-page-title'

    def prepare( self ):
        self.titles = {}

    def get_value( self , page_id , lang_id ):

        # Is it cached?
        if ( page_id , lang_id ) in self.titles:
            return self.titles[ ( page_id , lang_id ) ]

        # Look in the language-specific section
        xpath_expr = ( "/ws:data/ws:contents[ @lang = $lid ]/ws:pages"
                     + "/ws:page[ @id = $pid and @title ]" )
        r = self.site.get( xpath_expr , lid = lang_id , pid = page_id )
        if len( r ) >= 1:
            self.titles[ ( page_id , lang_id ) ] = r[ 0 ].attrib[ 'title' ]
            return self.titles[ ( page_id , lang_id ) ]

        # Look in the sitemap
        xpath_expr = "/ws:data/ws:sitemap/ws:page[ @id = $pid and @title ]"
        r = self.site.get( xpath_expr , pid = page_id )
        if len( r ) >= 1:
            self.titles[ ( page_id , lang_id ) ] = r[ 0 ].attrib[ 'title' ]
        else:
            self.titles[ ( page_id , lang_id ) ] = unicode( '(title not found)' )
        return self.titles[ ( page_id , lang_id ) ]



class GetPageAuthor( XPathFunction ):

    function = 'get-page-author'

    def prepare( self ):
        self.authors = {}

    def get_value( self , page_id , lang_id ):

        # Is it cached?
        if ( page_id , lang_id ) in self.authors:
            return self.authors[ ( page_id , lang_id ) ]

        # Look in the language-specific section
        xpath_expr = ( "/ws:data/ws:contents[ @lang = $lid ]/ws:pages"
                        + "/ws:page[ @id = $pid ]/ws:author" )
        r = self.site.get( xpath_expr , lid = lang_id , pid = page_id )
        if len( r ) >= 1:
            self.authors[ ( page_id , lang_id ) ] = r[ 0 ].text
            return self.authors[ ( page_id , lang_id ) ]

        # Look in the sitemap
        xpath_expr = "/ws:data/ws:sitemap/ws:page[ @id = $pid ]/ws:author"
        r = self.site.get( xpath_expr , pid = page_id )
        if len( r ) >= 1:
            self.authors[ ( page_id , lang_id ) ] = r[ 0 ].text
        else:
            # Use the site's main author, if any
            r = self.site.get( '/ws:data/ws:author' ) 
            if len( r ) >= 1:
                site_author = r[ 0 ].text
            else:
                site_author = ''
            self.authors[ ( page_id , lang_id ) ] = site_author
        return self.authors[ ( page_id , lang_id ) ]



class GetPageData( XPathFunction ):

    function = 'get-page-data'


    def prepare( self ):
        self.page_data = {}

    def get_node( self , page_id , lang_id ):

        # Is it cached?
        if ( page_id , lang_id ) in self.page_data:
            return self.page_data[ ( page_id , lang_id ) ]

        # Find the page's record
        xpath_expr = ( '/ws:data/ws:contents[ @lang = $lang ]'
                       + '/ws:pages/ws:page[ @id = $pid ]' )
        page_lst = self.site.get( xpath_expr , lang = lang_id , pid = page_id )
        if len( page_lst ) != 1:
            hdr = self.html( 'h2' , 'Error for page ' + page_id
                    + ' in language ' + lang_id )
            p = self.html( 'p' , 'The page could not be found '
                    + 'in the site\'s configuration.' )
            return [ hdr , p ]
        page = page_lst[ 0 ]

        # Does it have a separate data file?
        if 'datafile' not in page.attrib:
            # Nope, return the tag's contents
            self.page_data[ ( page_id , lang_id ) ] = list( page )
            return self.page_data[ ( page_id , lang_id ) ]

        # Get the parameters that determine the path to the page's data file
        file_path = self.site.path
        r = self.site.get( '/ws:data/ws:page-data' )
        if len( r ) == 1:
            file_path += r[ 0 ].text + '/'
        xpath_expr = '/ws:data/ws:contents[ @lang = $lang ]/@language-path'
        r = self.site.get( xpath_expr , lang = lang_id )
        if len( r ) == 1:
            file_path += lang_id + '/'
        else:
            xpath_expr = '/ws:data/ws:contents[ @lang = $lang ]/@base-path'
            r = self.site.get( xpath_expr , lang = lang_id )
            if len( r ) == 1:
                file_path += r[ 0 ].text + '/'
        file_path += page.attrib[ 'datafile' ] + '.xml'

        # Get the data from the file
        page_file = etree.parse( file_path )
        self.page_data[ ( page_id , lang_id ) ] = list( page_file.getroot() )
        return self.page_data[ ( page_id , lang_id ) ]



class GetSideBar( XPathFunction ):

    function = 'get-side-bar'

    def prepare( self ):
        self.sidebars = {}

    def get_node( self , page ):

        # Did we cache it?
        if page in self.sidebars:
            return self.sidebars[ page ]

        # Try to find the definition
        xpath_expr = "/ws:data/ws:side-bar[ @page = $pid ]"
        r = self.site.get( xpath_expr , pid = page )
        if len( r ) == 0:
            r = self.site.get( "/ws:data/ws:side-bar[ not( @page ) ]" )

        # Cache result and return
        if len( r ) == 0:
            self.sidebars[ page ] = [ ]
        else:
            self.sidebars[ page ] = [ r[ 0 ] ]
        return self.sidebars[ page ]



class GenerateBreadcrumb( XPathFunction ):

    function = 'generate-breadcrumb'

    def get_node( self , page_id , lang_id , omit_last , top_level_string ):

        # Find out the home page's identifier
        r = self.site.get( '/ws:data/ws:sitemap' )
        if 'home' in r[ 0 ].attrib:
            home_id = r[ 0 ].attrib[ 'home' ]
        else:
            home_id = 'main'

        if page_id == home_id:
            # This is the root, do we display the string?
            if top_level_string:
                span_text = self.func( GetI18NString , 'breadcrumb-top' ,
                        lang_id )
            else:
                span_text = self.func( GetPageTitle , page_id , lang_id )
            return [ self.html( 'span' , contents = span_text ) ]

        # Get the breadcrumb separator
        r = self.site.get( '/ws:data/ws:breadcrumb[ @separator ]' )
        if len( r ) >= 1:
            bc_sep = r[ 0 ].attrib[ 'separator' ]
        else:
            bc_sep = unicode( ' ' )

        # Not at the root
        bc = [ ]
        current = page_id
        previous = ''
        while True:

            # Create the crumb
            if not( current == page_id and omit_last ):
                n_text = self.func( GetPageTitle , current , lang_id )
                if previous != page_id or not( omit_last ):
                    if current == page_id:
                        span_text = n_text
                    else:
                        span_text = bc_sep
                    bc.insert( 0 , self.html( 'span' , span_text ) )

                if current != page_id:
                    # This is a link
                    link = self.ws( 'link-to' , n_text )
                    link.attrib[ 'page' ] = current
                    bc.insert( 0 , link )

            # End the loop if that was the home page
            if current == home_id:
                break

            # Get the previous page in the hierarchy
            previous = current
            xpath_expr = ( "/ws:data/ws:sitemap/ws:page"
                         + "[ @id = $pid and @parent ]" )
            r = self.site.get( xpath_expr , pid = current )
            if len( r ) >= 1:
                current = r[ 0 ].attrib[ 'parent' ]
            else:
                current = home_id

        return bc



class GetMetaData( XPathFunction ):

    function = 'get-meta-data'

    def prepare( self ):
        self.cache = {}

    def get_node( self , lang , page , tag , id_field = None ):
        # Check if the specified tag is supported
        if tag not in ( 'meta' , 'script' , 'css' , 'image-preload' ):
            raise ValueError( 'unknown tag "%s"' % tag )

        # Is the data cached?
        meta_id = ( tag , lang , page )
        if meta_id in self.cache:
            return self.cache[ meta_id ]

        # List of XPath expressions for the tag
        x_exprs = ( '' ,
                '/ws:contents[ @lang = $lang ]' ,
                '/ws:sitemap/ws:page[ @id = $page ]' ,
                '/ws:contents[ @lang = $lang ]/ws:pages/ws:page[ @id = $page ]'
            )

        # Depending on whether there is an id field or not, generate a list or
        # a dictionnary
        if id_field is None:
            data = [ ]
            values = [ ]
        else:
            data = { }

        # Iterate through the list of expressions
        for xe_base in x_exprs:
            xe = '/ws:data' + xe_base + ( '/ws:%s' % tag )
            r = self.site.get( xe , lang = lang , page = page )

            # Iterate through the results
            for t in r:

                if id_field is None:
                    # Handle lists
                    value = t.text
                    if value not in values:
                        data.append( deepcopy( t ) )
                        values.append( value )
                    continue

                if id_field not in t.attrib:
                    # Ignore tags that are missing their identifier
                    continue

                # Handle tags with identifiers
                data[ t.attrib[ id_field ] ] = deepcopy( t )

        # If there was an identifier, generate the list
        if id_field is not None:
            data = data.values( )

        # Cache the data and return it
        self.cache[ meta_id ] = data
        return data



class GetDataPath( XPathFunction ):

    function = 'get-data-uri'

    def prepare( self ):
        self.cache = {}

    def get_value( self , d_type ):

        if d_type not in ( 'js' , 'css' , 'img' ):
            raise ValueError( 'Invalid data type "%s"' % d_type )

        if d_type not in self.cache:
            r = self.site.get(
                    '/ws:data/ws:data-path/ws:path-to[ @type = $id ]' )
            if len( r ) >= 1:
                self.cache[ d_type ] = r[ 0 ].text
            else:
                self.cache[ d_type ] = d_type

        return self.cache[ d_type ]



class GetSiteTitle( XPathFunction ):

    function = 'get-site-title'

    def prepare( self ):
        self.cache = {}

    def get_value( self , lang ):

        if lang in self.cache:
            return self.cache[ lang ]

        r = self.site.get( '/ws:data/ws:contents[ @lang = $lang ]/ws:title' ,
                lang = lang )
        if len( r ) == 0:
            r = self.site.get( '/ws:data/ws:title' )
            if len( r ) == 0:
                self.cache[ lang ] = self.func( GetI18NString ,
                        'untitled-site' , lang )
        if lang not in self.cache:
            self.cache[ lang ] = r[ 0 ].text

        return self.cache[ lang ]




##############################################################################
# REQUEST HANDLER                                                            #
##############################################################################

def handler( req ):

    # Find the script's base URI and physical path
    script_uri , base_uri , data_path = get_paths( req )

    # Define the site, then register XSL functions
    site = SiteXML( data_path )
    registry = XPathRegistry( site )
    registry.add( GetI18NString , GetPageTitle , GetPageAuthor ,
            GetPageData , GetSideBar , GenerateBreadcrumb , GetMetaData ,
            GetDataPath , GetSiteTitle )

    # Find out which languages are supported
    languages = list( get_languages( site ) )

    # Extract requested language and page from the path
    r_lang = ""
    r_page = "main"
    if req.path_info:
        splitPath = req.path_info[ 1 : ].split( '/' )
        if len( splitPath ) == 1:
            r_page = splitPath.pop(0)
        elif len( splitPath ) >= 2:
            r_lang = splitPath.pop(0)
            r_page = splitPath.pop(0)

    # Check if the page name is valid
    if not re.match( '[a-z][a-z0-9\-]*$' , r_page , re.IGNORECASE ):
        r_page = "main"
        raise ValueError # FIXME: should be an error

    # Check if the selected language is valid
    if r_lang is not None and r_lang not in languages:
        r_lang = None

    # If there is no specified language, try to guess it
    # according to Accept-Language
    if r_lang is None:
        if req.headers_in.has_key( 'Accept-Language' ):
            accept_string = req.headers_in[ 'Accept-Language' ]
            r_lang = guess_language( accept_string , languages )
        else:
            r_lang = languages[ 0 ]

    # Load the XSLT stylesheet
    transform = etree.XSLT( etree.parse( data_path + 'site.xsl' ) )

    # Check if the client accepts application/xhtml+xml
    if req.headers_in.has_key( 'Accept' ):
        accept = re.split( '[;,]' , req.headers_in[ 'Accept' ].lower() )
        if 'application/xhtml+xml' in accept:
            req.content_type = 'application/xhtml+xml'
        else:
            req.content_type = 'text/html'
    else:
        req.content_type = 'text/html'

    req.write( str( transform( site.xml ,
            reqPage = "'" + r_page + "'" ,
            reqLang = "'" + r_lang + "'" ,
            webRoot = "'" + base_uri + "'" ,
            scriptPath = "'" + script_uri + "'" ) ) )

    return apache.OK
